/*
 * Copyright (c) 2006 Suguru ARAKAWA. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package jp.ac.hosei.k.cis.java2005.dita;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.MessageFormat;
import java.util.Arrays;

import jp.ac.hosei.k.cis.java2005.dita.swing.BinaryEditorSwingView;

/**
 * DITA - A Pure Java Simple Binary Editor.
 * @author arakawa
 * @version $Id: BinaryEditor.java,v 1.5 2005/11/17 06:23:12 java2005 Exp $
 */
public class BinaryEditor implements BinaryEditorController {

    private BinaryEditorView view;

    /**
     * プログラムエントリ。
     * @param args 一つ目の引数に開くファイルを設定してもよい
     */
    public static void main(String[] args) {
        if (args.length >= 2) {
            throw new IllegalArgumentException(Arrays.asList(args).toString());
        }

        BinaryEditor main = new BinaryEditor();
        BinaryEditorView view = new BinaryEditorSwingView(main);
        main.view = view;

        if (args.length == 1) {
            main.performFileOpen(new File(args[0]));
        }
    }

    /**
     * デフォルトコンストラクタ。
     */
    private BinaryEditor() {
        super();
    }

    /**
     * ファイルを開く処理を行う。
     */
    public void performFileOpen() {
        File selected = this.view.requestFileOpen();
        if (selected == null) {
            return;
        }
        performFileOpen(selected);
    }

    /**
     * ファイルを指定して、指定したファイルを開く処理を行う。
     * @param selected 開くファイル
     */
    private void performFileOpen(File selected) {
        if (selected == null) {
            throw new NullPointerException("selected"); //$NON-NLS-1$
        }

        try {
            byte[] data = readFile(selected);
            this.view.setData(data);
            this.view.notifyFileOpened(selected);
        }
        catch (FileNotFoundException e) {
            this.view.notifyErrorOccurred(
                Messages.getString("BinaryEditor.file_not_found_title"), //$NON-NLS-1$
                MessageFormat.format(
                    Messages.getString("BinaryEditor.file_not_found_message"), //$NON-NLS-1$
                    new Object[]{selected.getAbsolutePath()})
                );
        }
        catch (IOException e) {
            this.view.notifyErrorOccurred(
                Messages.getString("BinaryEditor.read_error_title"), //$NON-NLS-1$
                MessageFormat.format(
                    Messages.getString("BinaryEditor.read_error_message"), //$NON-NLS-1$
                    new Object[]{selected.getAbsolutePath()})
                );
        }
    }

    /**
     * ファイルを閉じる処理を行う。
     */
    public void performFileSave() {
        File selected = this.view.requestFileSave();
        if (selected == null) {
            return;
        }
        try {
            byte[] data = this.view.getData();
            writeFile(selected, data);
            this.view.notifyFileSaved(selected);
        }
        catch (IOException e) {
            this.view.notifyErrorOccurred(
                Messages.getString("BinaryEditor.write_error_title"), //$NON-NLS-1$
                MessageFormat.format(
                    Messages.getString("BinaryEditor.write_error_message"), //$NON-NLS-1$
                    new Object[]{selected.getAbsolutePath()})
                );
        }
    }

    /**
     * プログラムを終了する処理を行う。
     */
    public void performExit() {
        this.view.requestExit();
    }

    /**
     * 実際にファイルからデータを読み込む処理。
     * @param file 読み込み元のファイル
     * @return 読み込んだデータ
     * @throws FileNotFoundException 読み込もうとしたファイルが見つからなかった場合
     * @throws IOException ファイル読み込み中に例外が発生した場合
     * @throws NullPointerException 引数に<code>null</code>が含まれていた場合
     */
    private byte[] readFile(File file) throws FileNotFoundException, IOException {
        if (file == null) {
            throw new NullPointerException("file"); //$NON-NLS-1$
        }
        
        ByteArrayOutputStream data = new ByteArrayOutputStream();
        InputStream is = new BufferedInputStream(new FileInputStream(file));
        try {
            byte[] buf = new byte[256];
            while (true) {
                int read = is.read(buf);
                if (read == -1) {
                    break;
                }
                data.write(buf, 0, read);
            }
        }
        finally {
            is.close();
        }
        data.flush();
        return data.toByteArray();
    }

    /**
     * 実際にファイルにデータを書き込む処理。
     * @param file 書き込み先のファイル
     * @param data 書き込むデータ
     * @throws IOException ファイル書き込み中に例外が発生した場合
     * @throws NullPointerException 引数に<code>null</code>が含まれていた場合
     */
    private void writeFile(File file, byte[] data) throws IOException {
        if (file == null) {
            throw new NullPointerException("file"); //$NON-NLS-1$
        }
        if (data == null) {
            throw new NullPointerException("data"); //$NON-NLS-1$
        }
        
        OutputStream os = new BufferedOutputStream(new FileOutputStream(file));
        try {
            os.write(data);
        }
        finally {
            os.close();
        }
    }
}
