/*
 * Copyright (c) 2006 Suguru ARAKAWA. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package jp.ac.hosei.k.cis.java2005.dita.swing;

import javax.swing.table.AbstractTableModel;

/**
 * バイナリエディタのテーブルモデル。
 * @author arakawa
 * @version $Id: BinaryEditorTableModel.java,v 1.2 2005/11/17 05:23:10 java2005 Exp $
 */
public class BinaryEditorTableModel extends AbstractTableModel {

    private static final long serialVersionUID = 1L;

    private static final int ROW_SIZE = 16;

    private static final String COL_NAME_ADDRESS = Messages.getString("BinaryEditorTableModel.address"); //$NON-NLS-1$
    
    private static final String EOF = "--"; //$NON-NLS-1$
    
    private static final Integer[] BYTES;
    static {
        BYTES = new Integer[256];
        for (int i = 0; i < 256; i++) {
            BYTES[i] = new Integer(i);
        }
    }

    private static final String[] COL_NAMES;
    static {
        COL_NAMES = new String[ROW_SIZE + 1];
        COL_NAMES[0] = COL_NAME_ADDRESS;
        for (int i = 1; i < COL_NAMES.length; i++) {
            COL_NAMES[i] = "+" + toHexString(i - 1); //$NON-NLS-1$
        }
    }
    
    private static final int ACTION_EOF = -1;
    private static final int ACTION_DELETE = -2;
    private static final int ACTION_CALM = -3;
    
    private byte[] contents;
    private int size;

    /**
     * 空のデータを持つテーブルモデルを作成する。
     */
    public BinaryEditorTableModel() {
        this(new byte[0]);
    }

    /**
     * 初期値を指定してテーブルモデルを作成する。
     * @param contents 初期値
     */
    public BinaryEditorTableModel(byte[] contents) {
        super();
        if (contents == null) {
            throw new NullPointerException("contents"); //$NON-NLS-1$
        }
        // 256バイト余計に取得
        this.contents = expandByteArray(contents, 256);
    }
    
    /**
     * 指定したデータに内容を設定する。
     * @param data データ
     */
    public void setData(byte[] data) {
        if (data == null) {
            throw new NullPointerException("contents"); //$NON-NLS-1$
        }
        // 256バイト余計に取得
        this.contents = expandByteArray(data, 256);
        this.size = data.length;
        fireTableDataChanged();
    }
    
    /**
     * 編集中のデータを取得する。
     * @return データ
     */
    public byte[] getData() {
        byte[] data = new byte[this.size];
        System.arraycopy(this.contents, 0, data, 0, data.length);
        return data;
    }
    
    /**
     * カラムの型を返す。
     * @param columnIndex カラム
     * @return カラムの型
     */
    public Class getColumnClass(int columnIndex) {
        return Object.class;
    }

    /**
     * 列の名前を取得する。
     * @param column 名前を取得する列
     * @return 列の名前
     */
    public String getColumnName(int column) {
        return COL_NAMES[column];
    }

    /**
     * 指定した位置のセルが編集可能かどうか調べる。
     * @param rowIndex 行
     * @param columnIndex 無視される
     * @return <code>rowIndex == 0</code> 以外は常に <code>true</code>
     */
    public boolean isCellEditable(int rowIndex, int columnIndex) {
        return columnIndex != 0;
    }

    /**
     * 指定した位置の値を変更する。
     * @param aValue 新しい値
     * @param rowIndex 値を変更する位置の行
     * @param columnIndex 値を変更する位置の列
     */
    public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
        if (aValue == null || !(aValue instanceof String)) {
            // do nothing
            return;
        }
        if (columnIndex == 0) {
            // address is not editable..
            return;
        }
        
        // parse
        int n;
        try {
            n = parseInputValue((String) aValue);
        }
        catch (NumberFormatException e) {
            // do nothing
            return;
        }
        
        int offset = positionToOffet(rowIndex, columnIndex);
        
        if (n == ACTION_DELETE) {
            deleteContentAt(offset);
            fireTableDataChanged();
            return;
        }
        else if (n == ACTION_EOF) {
            // MEMO EOF を指定しても何もしない
            return;
        }
        else if (n == ACTION_CALM) {
            // 何もしないコマンド
            return;
        }
        else {
            // 実際に書き込む
            prepareBuffere(offset);
            if (n == getContentAsInt(offset)) {
                // 過去と同じなら何もしない
                return;
            }
            this.contents[offset] = (byte) n;
            fireTableDataChanged();
        }
    }

    /**
     * 入力されたバイトデータ文字列を解析する。
     * @param input 解析する文字列
     * @return 解析した結果の値
     * @throws NumberFormatException 不正な値が入力された場合
     */
    private int parseInputValue(String input) {
        String trimmed = input.trim();
        if (trimmed.length() == 0) {
            return ACTION_DELETE;
        }
        if (trimmed.equals(EOF)) {
            return ACTION_EOF;
        }
        int n = Integer.valueOf(trimmed, 16).intValue();
        if (n < 0 || n >= 256) {
            return ACTION_CALM;
        }
        else {
            return n;
        }
    }

    /**
     * バッファを書き込み用に用意する。
     * バッファサイズが十分でなければ拡張する。
     * @param offset 書き込み先
     */
    private void prepareBuffere(int offset) {
        // check buffer
        if (offset >= this.size) {
            // ファイルの終端以降に書き込んだ
            if (offset >= this.contents.length) {
                // バッファの範囲外に書き込んだ
                this.contents = expandByteArray(this.contents, (offset - this.contents.length) + 256);
            }
            // 新しく作った部分を 0 埋め
            for (int i = this.size; i < offset; i++) {
                this.contents[i] = 0;
            }
            this.size = offset + 1;
        }
    }
    
    /**
     * 行数を取得する。
     * @return 行数
     */
    public int getRowCount() {
        return this.size / ROW_SIZE + 1;
    }

    /**
     * 列数を取得する。
     * @return 列数
     */
    public int getColumnCount() {
        return ROW_SIZE + 1;
    }

    /**
     * 指定した位置の値を取得する。
     * @param rowIndex 取得する位置の行
     * @param columnIndex 取得する位置の列
     * @return 取得した値
     * @throws IllegalArgumentException 範囲外の列が指定された場合
     */
    public Object getValueAt(int rowIndex, int columnIndex) {
        // アドレス表示
        if (columnIndex == 0) {
            return new Integer(rowIndex * ROW_SIZE);
        }
        // 値表示
        else if (columnIndex < ROW_SIZE + 1) {
            int value = getContentAsInt(positionToOffet(rowIndex, columnIndex));
            if (value != -1) {
                return BYTES[value];
            }
            else {
                return null;
            }
        }
        else {
            throw new IllegalArgumentException("columnIndex="+columnIndex); //$NON-NLS-1$
        }
    }
    
    /**
     * <code>0</code> 以上 <code>255</code> 以下の値を二桁の16進数を表す文字列に変換する。
     * @param n 変換する値
     * @return 変換した値
     * @throws IllegalArgumentException <code>0</code> 以上 <code>255</code> 以下の値が入力されなかった場合
     */
    private static String toHexString(int n) {
        if (n < 0 || n >= 256) {
            throw new IllegalArgumentException("n=" + n); //$NON-NLS-1$
        }
        return Integer.toHexString((n >> 4) & 0x0f) + Integer.toHexString(n & 0x0f);
    }
    
    /**
     * <code>byte</code> 配列の長さを拡張する。
     * 拡張した配列は、元の配列の内容がコピーされる。
     * @param original 元の配列
     * @param delta 増やす容量
     * @return 拡張した配列
     * @throws NullPointerException <code>original</code> に <code>null</code> を指定した場合
     * @throws IllegalArgumentException <code>delta</code> に負の値を指定した場合
     */
    private static byte[] expandByteArray(byte[] original, int delta) {
        if (original == null) {
            throw new NullPointerException("original"); //$NON-NLS-1$
        }
        if (delta < 0) {
            throw new IllegalArgumentException("delta=" + delta); //$NON-NLS-1$
        }
        byte[] result = new byte[original.length + delta];
        System.arraycopy(original, 0, result, 0, original.length);
        return result;
    }
    
    /**
     * 指定したテーブル内での位置を、内部バッファの位置に変換する。
     * @param row テーブル内での位置 (行)
     * @param col テーブル内での位置 (列)
     * @return 内部バッファの位置
     * @throws IllegalArgumentException 内部バッファの位置に変換できない場合 (col = 0 など)
     */
    private static int positionToOffet(int row, int col) {
        if (col < 1) {
            throw new IllegalArgumentException("col=" + col); //$NON-NLS-1$
        }
        return row * ROW_SIZE + col - 1;
    }

    /**
     * 内部バッファから指定した位置の値を取得する。
     * 内部バッファの範囲外の値を指定した場合は、<code>-1</code> を返す。
     * @param offset 内部バッファの位置
     * @return 指定した位置の値、または <code>-1</code>
     * @throws IllegalArgumentException オフセットに負の値を指定した場合
     */
    private int getContentAsInt(int offset) {
        if (offset < 0) {
            throw new IllegalArgumentException("offset=" + offset); //$NON-NLS-1$
        }
        if (offset >= this.size) {
            return -1;
        }
        return this.contents[offset] & 0xff;
    }
    
    /**
     * 内部バッファから指定した位置の値を削除する。
     * 範囲外の値を指定した場合は削除を行わない。
     * @param offset 内部バッファの位置
     * @throws IllegalArgumentException オフセットに負の値を指定した場合
     */
    private void deleteContentAt(int offset) {
        if (offset < 0) {
            throw new IllegalArgumentException("offset=" + offset); //$NON-NLS-1$
        }
        if (offset >= this.size) {
            return;
        }
        for(int i = offset; i < this.size - 1; i++) {
            this.contents[i] = this.contents[i + 1];
        }
        
        this.size--;
    }
}
